> module Decks(Flux(..), FEDeck(),
>              empty, singletonF, singletonE, ecat, cat,
>              prefix, suffix, cons, snoc, uncons, cons', isSingletonE,
>              feFold, feMap, feFoldr, feFoldl, feFoldE, feBinFoldE,
>              eLength
>             ) where

> infixr 5 ~~


> class Flux f where
>   emptyFlux :: f
>   fluxIsEmpty :: f -> Bool
>   (~~) :: f -> f -> f

> instance Flux [a] where
>   emptyFlux = []
>   fluxIsEmpty = null
>   (~~) = (++)

> data FEDeck f e = Sing f
>                 | Head f e (FEDeck f e)
>                 | Tail (FEDeck f e) e f
>                 | Cat (FEDeck f e) (FEDeck f e)
>                 | ECat (FEDeck f e) e (FEDeck f e)

> singletonF :: a -> FEDeck a b
> singletonF x = Sing x

> empty :: Flux a => FEDeck a b
> empty = Sing emptyFlux

> singletonE :: Flux a => b -> FEDeck a b
> singletonE e = Head emptyFlux e (Sing emptyFlux)

> ecat :: FEDeck a b -> b -> FEDeck a b -> FEDeck a b
> ecat = ECat

> cons :: a -> b -> FEDeck a b -> FEDeck a b
> cons = Head

> snoc :: FEDeck a b -> b -> a -> FEDeck a b
> snoc = Tail

> cons' :: Flux a => a -> Maybe (b,FEDeck a b) -> FEDeck a b
> cons' x Nothing = Sing x
> cons' x (Just (e,d)) = Head x e d

> uncons :: Flux a => FEDeck a b -> (a,Maybe (b,FEDeck a b))
> uncons (Sing y) = (y,Nothing)
> uncons (Head y e r) = (y,Just (e,r))
> uncons (Tail d e y) = let (x,r) = uncons d
>                           j = case r of
>                                 Nothing -> (e,Sing y)
>                                 Just (e',r') -> (e', Tail r' e y)
>                       in (x, Just j)
> uncons (Cat d1 d2) = let (x,r) = uncons d1
>   in case r of
>        Nothing -> let (x',r') = uncons d2
>                   in (x ~~ x', r')
>        Just (e,r') -> (x, Just (e, Cat r' d2))
> uncons (ECat d1 e d2) = let (x,r) = uncons d1
>   in case r of
>        Nothing -> (x,Just (e,d2))
>        Just (e',r') -> (x, Just (e', ECat r' e d2))

> isSingletonE :: Flux a => FEDeck a b -> Maybe b
> isSingletonE d = let (f,m) = uncons d in
>   case m of Nothing -> Nothing
>             Just (e,d') -> if fluxIsEmpty f && isEmpty d'
>                            then Just e else Nothing

> prefix :: Flux a => a -> FEDeck a b -> FEDeck a b
> prefix x (Sing y) = Sing (x ~~ y)
> prefix x (Head y e r) = Head (x ~~ y) e r
> prefix x (Tail d e y) = Tail (prefix x d) e y
> prefix x (Cat d1 d2) = Cat (prefix x d1) d2
> prefix x (ECat d1 e d2) = ECat (prefix x d1) e d2

> suffix  :: Flux a => FEDeck a b -> a -> FEDeck a b
> suffix (Sing y) x = Sing (y ~~ x)
> suffix (Head y e r) x = Head y e (suffix r x)
> suffix (Tail d e y) x = Tail d e (y ~~ x)
> suffix (Cat d1 d2) x = Cat d1 (suffix d2 x)
> suffix (ECat d1 e d2) x = ECat d1 e (suffix d2 x)

> feFold :: Flux a => (r -> b -> r -> r) -> (a -> r) -> FEDeck a b -> r
> feFold g f = f' where
>   f' (Sing x) = f x
>   f' (Head x e d) = g (f x) e (f' d)
>   f' (Tail d e x) = g (f' d) e (f x)
>   f' (Cat d1 d2) = f' (cat d1 d2)
>   f' (ECat d1 e d2) = g (f' d1) e (f' d2)

> feFoldE :: Flux a => (r -> b -> r -> r) -> r -> FEDeck a b -> r
> feFoldE g r = f' where
>   f' (Sing x) = r
>   f' (Head x e d) = g r e (f' d)
>   f' (Tail d e x) = g (f' d) e r
>   f' (Cat d1 d2) = f' (cat d1 d2)
>   f' (ECat d1 e d2) = g (f' d1) e (f' d2)

> feBinFoldE :: Flux a => (r -> r -> r) -> (b -> r) -> r -> FEDeck a b -> r
> feBinFoldE g s r = f' where
>   f' (Sing x) = r
>   f' (Head x e d) = g (s e) (f' d)
>   f' (Tail d e x) = g (f' d) (s e)
>   f' (Cat d1 d2) = f' (cat d1 d2)
>   f' (ECat d1 e d2) = (f' d1) `g` ((s e) `g` (f' d2))

> feMap :: Flux a => (a -> b) -> (c -> d) -> FEDeck a c -> FEDeck b d
> feMap f g = f' where
>   f' (Sing x) = Sing (f x)
>   f' (Head x e d) = Head (f x) (g e) (f' d)
>   f' (Tail d e x) = Tail (f' d) (g e) (f x)
>   f' (Cat d1 d2) = f' (cat d1 d2)
>   f' (ECat d1 e d2) = ECat (f' d1) (g e) (f' d2)

> cat :: Flux a => FEDeck a b -> FEDeck a b -> FEDeck a b
> cat (Sing x)       d2 = prefix x  d2
> cat (Head x  e d1) d2 = Head x  e (Cat d1 d2)
> cat (Tail d1 e x ) d2 = ECat d1 e (prefix x d2)
> cat (Cat  d1   d2) d3 = cat  d1   (Cat d2 d3)
> cat (ECat d1 e d2) d3 = ECat d1 e (Cat d2 d3)

> feFoldr ::Flux a =>  (a -> b -> r -> r) -> (a -> r) -> FEDeck a b -> r
> feFoldr g = h where
>   h f (Sing x)       = f x
>   h f (Head x e d)   = g x e (h f d)
>   h f (Tail d e x)   = h f' d  where f' y = g y e (f x)
>   h f (Cat d1 d2)    = h f' d1 where f' x = h f (prefix x d2)
>   h f (ECat d1 e d2) = h f' d1 where f' x = g x e (h f d2)

> feFoldl :: Flux a => (r -> b -> a -> r) -> (a -> r) -> FEDeck a b -> r
> feFoldl g = h where
>   h f (Sing x)       = f x
>   h f (Head x e d)   = h (g (f x) e) d
>   h f (Tail d e x)   = g (h f d) e x
>   h f (Cat d1 d2)    = h f' d2 where f' x = h f (suffix d1 x)
>   h f (ECat d1 e d2) = h f' d2 where f' x = g (h f d1) e x

> isEmpty :: Flux a => FEDeck a b -> Bool
> isEmpty (Sing f) = fluxIsEmpty f
> isEmpty (Cat d1 d2) = isEmpty d1 && isEmpty d2
> isEmpty _ = False

> instance Flux a => Flux (FEDeck a b) where
>   emptyFlux = Sing emptyFlux
>   fluxIsEmpty = isEmpty
>   (~~) = cat

> eLength :: FEDeck f e -> Int
> eLength (Sing f) = 0
> eLength (Head f e d) = 1 + eLength d
> eLength (Tail d e f) = 1 + eLength d
> eLength (Cat d1 d2) = eLength d1 + eLength d2
> eLength (ECat d1 e d2) = 1 + eLength d1 + eLength d2
