\documentclass{article}

%{{{ packages
\usepackage{verbatim}
%
\makeatletter
\def\itverb{%
\def\verbatim@font{\ttfamily\slshape
                   \let\do\do@noligs
                   \verbatim@nolig@list}%
}%itverb
\makeatother
%
\newenvironment{code}{\small\verbatim}{\endverbatim\normalsize}
\newenvironment{pseudocode}{\small\it\verbatim}{\endverbatim\normalsize}
\newenvironment{session}{\small\itverb\verbatim}{\endverbatim\normalsize}

\usepackage{latexsym,amssymb}
\usepackage[dvips]{graphicx}
%}}}

\author{{\sc Wolfram Kahl}}
\title{ICFP Contest 2001}

%{{{ Settings
\topmargin 0pt
\setlength{\voffset}{-15mm}
\setlength{\hoffset}{-9mm}
\setlength{\oddsidemargin}{0mm}
\setlength{\evensidemargin}{-5.5mm}
\setlength{\textwidth}{172mm}
\setlength{\textheight}{253mm}

\pagestyle{headings}
%%\pagestyle{empty}

\parindent0pt
\parskip4pt
%}}}

\begin{document}
\maketitle

\tableofcontents

%{{{ \datespecial
\edef\dateps{{! userdict begin /bop-hook {gsave 240 30 translate
/Helvetic-Narrow findfont 9 scalefont setfont
0 0 moveto (\number\year-\number\month-\number\day) show
10 0 rmoveto (\number\time) show
30 0 rmoveto
dup 1 add 4 string cvs show
grestore} def end}}
\def\datespecial{\expandafter\special\dateps}
\datespecial
%}}}

\centerline{\texttt{http://cristal.inria.fr/ICFP2001/prog-contest/task.ps}}

\begin{code}
import qualified Char
import qualified List
import System
import Random
import Decks
import Parsing
import STFun
\end{code}

%{{{ \section{Data Structure}
\section{Data Structure}

%{{{ data Tag
\begin{code}
type Colour = Char
colours = "rgbcmykw"

type UL = Int

type Size = Int
sizes :: [Size]
sizes = [0 .. 9]

data Tag = U
         | Col ! Colour
         | Size ! Size
         | B
         | EM
         | I
         | PL
         | S
         | TT
  deriving (Eq,Ord)

tags = [U,B,EM,I,PL,S,TT] ++ map Col colours ++ map Size sizes

asc = filter (`notElem` "<>") [Char.chr 33 .. Char.chr 126]
white = " \n\r\t"
chars = asc ++ white
chars_len = length chars - 1

instance Show Tag where
  showsPrec _ U = ('U' :)
  showsPrec _ (Col c) = (c :)
  showsPrec _ (Size i) = shows i
  showsPrec _ B = ('B' :)
  showsPrec _ EM = ('E' :) . ('M' :)
  showsPrec _ I = ('I' :)
  showsPrec _ PL = ('P' :) . ('L' :)
  showsPrec _ S = ('S' :)
  showsPrec _ TT = ('T' :) . ('T' :)

instance Read Tag where
  readsPrec _ [] = []
  readsPrec _ (c : cs) =
   case c of
    'B' -> [(B,cs)]
    'I' -> [(I,cs)]
    'S' -> [(S,cs)]
    'U' -> [(U,cs)]
    '0' -> [(Size 0,cs)]
    '1' -> [(Size 1,cs)]
    '2' -> [(Size 2,cs)]
    '3' -> [(Size 3,cs)]
    '4' -> [(Size 4,cs)]
    '5' -> [(Size 5,cs)]
    '6' -> [(Size 6,cs)]
    '7' -> [(Size 7,cs)]
    '8' -> [(Size 8,cs)]
    '9' -> [(Size 9,cs)]
    'r' -> [(Col c,cs)]
    'g' -> [(Col c,cs)]
    'b' -> [(Col c,cs)]
    'c' -> [(Col c,cs)]
    'm' -> [(Col c,cs)]
    'y' -> [(Col c,cs)]
    'k' -> [(Col c,cs)]
    'w' -> [(Col c,cs)]
    'E' -> case cs of
            ('M' : cs') -> [(EM,cs')]
            _ -> []
    'P' -> case cs of
            ('L' : cs') -> [(PL,cs')]
            _ -> []
    'T' -> case cs of
            ('T' : cs') -> [(TT,cs')]
            _ -> []
    _   -> []
\end{code}
%}}}

%{{{ data Doc
\begin{code}
tagStart tag = ('<' :) . shows tag . ('>' :)
tagEnd   tag = ('<' :) . ('/' :) . shows tag . ('>' :)

data Doc = Doc (FEDeck String (Tag,Doc))

instance Show Doc where
  showsPrec i (Doc d) = feFold fe ff d
    where fe r1 (tag,d') r2 = r1 . tagStart tag . showsPrec i d' .
                                   tagEnd   tag . r2
          ff s = (s ++)

readsTag :: ReadS (Bool, Tag)
readsTag ('<' : s) =
  case s of
    ('/' : s1) -> readsTag' False s1
    _          -> readsTag' True  s
readsTag _ = []


readsTag' b s = case reads $ dropWhile Char.isSpace s of
  ((tag, s1) : _) -> case dropWhile ('>' /=) s1 of
                       _ : s2 -> [((b,tag),s2)]
                       _ -> []
  _ -> []

readsText s = [span ('<' /=) s]

readsStartTag :: ReadS Tag
readsStartTag = readsTag <?>
  (\ (b,tag) -> if b then Just tag else Nothing)

readsEndTag :: Tag -> ReadS ()
readsEndTag tag = readsTag <?>
  (\ (b,tag) -> if b then Nothing else Just ())

readsTagDoc :: ReadS (Tag,Doc)
readsTagDoc = readsStartTag `pThen` \ tag ->
              readsDoc `pThen` \ doc ->
              readsEndTag tag `pThen` \ _ -> pSucceed (tag,doc)

readsDoc :: ReadS Doc
readsDoc = readsText `pThen` \ f ->
           ((readsTagDoc `pThen` \ p ->
             readsDoc `pThen` \ (Doc d) ->
             pSucceed (Doc $ Decks.cons f p d)
            )
            <||> pSucceed (Doc $ singletonF f))

instance Read Doc where
  readsPrec _ = readsDoc
\end{code}
%}}}

%{{{ generateRandomDoc
\begin{code}
swapPair (x,y) = (y,x)

randomRSTF :: (Random a, RandomGen g) => (a, a) -> STFun g a
randomRSTF bds = STFun (swapPair . randomR bds)

rChar :: RandomGen g => STFun g Char
rChar = do i <- randomRSTF (0, chars_len)
           return (chars !! i)

rLen :: RandomGen g => Int -> STFun g Int
rLen i = randomRSTF (0,i)

rText :: RandomGen g => STFun g String
rText = do len <- rLen 4
           mapM (const rChar) [1 .. len]

rTag :: RandomGen g => STFun g Tag
rTag = do i <- rLen 24
          return (tags !! i)

rDeck w n = do t <- rText
               if n < 1 || w < 1
                then return (singletonF t)
                else do d <- rDoc (n - 1)
                        tag <- rTag
                        r <- rDeck (w - 1) n
                        return (Decks.cons t (tag,d) r)

rDoc n = do depth <- rLen n
            width <- rLen 10
            d <- rDeck width n
            return $ Doc d

generateRandomDoc =
 do args <- getArgs
    let n = case args of
              [] -> 6
              (a : _) -> case reads a of
                 ((i,_):_) -> i :: Int
                 _ -> 6
    rand <- newStdGen
    print $ runSTFun (rDoc n) rand
\end{code}
%}}}

%}}}

%{{{ \section{Optimize}
\section{Optimize}

%{{{ data Context
\begin{code}
data Context = Context
  {c_b    :: ! Bool
  ,c_em   :: ! Bool
  ,c_i    :: ! Bool
  ,c_s    :: ! Bool
  ,c_tt   :: ! Bool
  ,c_u    :: ! Int
  ,c_size :: ! (Maybe Int)
  ,c_col  :: ! (Maybe Char)
  } deriving Eq

c_init = Context
  {c_b = False
  ,c_em = False
  ,c_i = False
  ,c_s = False
  ,c_tt = False
  ,c_u = 0
  ,c_size = Nothing
  ,c_col = Nothing
  }
\end{code}
%}}}

\verb|c_tag tag c| is the context inside a \verb|tag|-element
if \verb|c| is the context outside.

%{{{ c_tag
\begin{code}
c_tag :: Tag -> Context -> Context
c_tag B  c = c {c_b = True} 
c_tag EM c = if c_s c then c
             else c {c_em = not $ c_em c} 
c_tag I  c = c {c_i = True}
c_tag PL c = c {c_s  = False
               ,c_em = False
               ,c_i  = False
               ,c_b  = False
               ,c_tt = False
               ,c_u  = 0
               }
c_tag S  c = c {c_s = True, c_em = False}
c_tag TT c = c {c_tt = True}
c_tag U  c = let u = c_u c
             in if u >= 3 then c else c {c_u = u + 1}
c_tag (Size i) c = c {c_size = Just i}
c_tag (Col co) c = c {c_col  = Just co}
\end{code}
%}}}

\verb|c_implies c tag| holds iff \verb|c_tag tag c == c|.

\begin{code}
c_implies c tag = c_tag tag c == c
\end{code}

\begin{code}
isWSpace c = c `elem` white

wCompress1 [] = []
wCompress1 (c:cs) | isWSpace c  =  c : wCompress1 (dropWhile isWSpace cs)
                  | otherwise   =  c : wCompress1 cs
\end{code}

%{{{ opt1doc
\begin{code}
opt1doc c (Doc d) = Doc $ opt1deck c d
opt1deck c = feFold fe ff
 where
  ff = singletonF . (if c_tt c then id else wCompress1)
  fe d1 (tag, doc@(Doc d)) d2 =
       if c_implies c tag
       then d1 `cat` (opt1deck c d `cat` d2)
       else ecat d1 (tag, opt1doc (c_tag tag c) doc) d2
\end{code}
%}}}

%{{{ data Props
\begin{code}
data Props = Props
  {p_b    :: ! Bool
  ,p_em   :: ! Bool
  ,p_i    :: ! Bool
  ,p_s    :: ! Bool
  ,p_tt   :: ! Bool
  ,p_u    :: ! Bool
  ,p_size :: ! Bool
  ,p_col  :: ! Bool
  } deriving Eq
\end{code}
%}}}

%{{{ relevant
If \verb|p| contains the properties used by a some document
contained in \verb|tag| inside context \verb|c|,
then \verb|relevant c p tag| indicates whether \verb|tag| is relevant
for the included document.
\begin{code}
relevant c p U = if c_u c >= 3 then False else p_u p
relevant c p (Col col)   = case c_col c of
  Nothing -> p_col  p
  Just col' -> if col' == col then False else p_col  p
relevant c p (Size size) = case c_size c of
  Nothing -> p_size p
  Just size' -> if size' == size then False else p_size p
relevant c p B  = if c_b c then False else p_b p
relevant c p EM = if c_s c then False else p_em p
relevant c p I  = if c_i c then False else p_i p
relevant c p PL = (c_s  c && p_s  p) ||
                  (c_em c && p_em p) ||
                  (c_i  c && p_i  p) ||
                  (c_b  c && p_b  p) ||
                  (c_tt c && p_tt p) ||
                  (c_u c /= 0 && p_u p)
relevant c p S  = if c_s c then False else p_s p
relevant c p TT = if c_tt c then False else p_tt p
\end{code}
%}}}

%{{{ noProps, allProps, whiteProps, opProps
\begin{code}
noProps = Props
  {p_b    = False
  ,p_em   = False
  ,p_i    = False
  ,p_s    = False
  ,p_tt   = False
  ,p_u    = False
  ,p_size = False
  ,p_col  = False
  }

allProps = Props
  {p_b    = True
  ,p_em   = True
  ,p_i    = True
  ,p_s    = True
  ,p_tt   = True
  ,p_u    = True
  ,p_size = True
  ,p_col  = True
  }

whiteProps c = Props
  {p_b    = False
  ,p_em   = False
  ,p_i    = False
  ,p_s    = False
  ,p_tt   = True
  ,p_u    = True
  ,p_size = True
  ,p_col  = c_u c /= 0
  }

opProps op (Props a1 b1 c1 d1 e1 f1 g1 h1)
           (Props a2 b2 c2 d2 e2 f2 g2 h2) =
  Props (a1 `op` a2)
        (b1 `op` b2)
        (c1 `op` c2)
        (d1 `op` d2)
        (e1 `op` e2)
        (f1 `op` f2)
        (g1 `op` g2)
        (h1 `op` h2)

-- orProps = opProps (||)
orProps a b = if a == allProps then a else opProps (||) a b
\end{code}
%}}}

%{{{ usedProps
If \verb|p| contains the properties used by a some document
contained in \verb|tag| inside context \verb|c|,
than \verb|usedProps c p tag| contains those properties used by
\verb|(tag,doc)| in \verb|c|.
\begin{code}
usedProps c p U = p
usedProps c p (Col col) = p {p_col = False}
usedProps c p (Size s) = p {p_size = False}
usedProps c p B = p {p_b = False}
usedProps c p EM = if c_s c then p {p_em = False} else p
usedProps c p I = p {p_i = False}
usedProps c p PL = p {p_s  = False
                     ,p_em = False
                     ,p_i  = False
                     ,p_b  = False
                     ,p_tt = False
                     ,p_u  = False
                     }
usedProps c p S  = p {p_s = False, p_em = False}
usedProps c p TT = p {p_tt = False}
\end{code}
%}}}

\begin{code}
pupd1 f (x,y) = (f x, y)

opt2doc c (Doc d) = pupd1 Doc $ opt2deck c d

textProps c [] = noProps
textProps c s = if all isWSpace s
                then whiteProps c
                else allProps

opt2deck c = feFold fe ff
 where
  ff0 = singletonF . (if c_tt c then id else wCompress1)
  ff s = (ff0 s, textProps c s)
  fe (d1,p1) (tag,doc@(Doc d)) (d2,p2) = let
         c' = c_tag tag c
         (doc',p) = opt2doc c' doc
    in if c_implies c tag || not (relevant c p tag)
       then let (d',p) = opt2deck c d
            in (d1 `cat` (d' `cat` d2), p1 `orProps` (p `orProps` p2))
       else let p' = usedProps c p tag
            in (ecat d1 (tag,doc') d2, p1 `orProps` (p' `orProps` p2))
\end{code}

\begin{code}
deckContract c d = let
  (t,m) = uncons d
 in case m of
   Nothing -> (t,Nothing)
   Just (p, d1) -> deckContract1 c t p d1

deckContract1 c t (tag,doc@(Doc d')) d1 = let
     (t1,m1) = deckContract c d1
     (t1w,t1rest) = span isWSpace t1
     c' = c_tag tag c
     move = not (relevant c (whiteProps c) tag)
     (d'',t1') =
       if not (null t1w) && move
       then (suffix d' t1w, t1rest)
       else (d', t1)
     (t',d''') = let (dt,dm) = deckContract c' d''
                     (dtw,dtrest) = span isWSpace dt
                 in if not (null dtw) && move
                    then (t ++ dtw, cons' dtrest dm)
                    else (t,cons' dt dm)
     p' = (tag,Doc d''')
   in case m1 of
        Nothing -> (t',Just (p',singletonF t1'))
        Just (p''@(tag1,doc1@(Doc d1')),d2) ->
          if null t1' && tag == tag1
          then deckContract1 c t (tag,Doc (d''' `cat` d1')) d2
          else (t',Just (p', cons' t1' m1))

contract1 (Doc d) = Doc $ uncurry cons' $ deckContract c_init d
\end{code}


\begin{code}
data XDoc = XDoc (FEDeck String [([Tag],XDoc)])
type XDeck = FEDeck String [([Tag],XDoc)]

docToXDoc (Doc d) = XDoc (deckToXDoc d)
deckToXDoc = uncurry cons' . deckToXDoc0
deckToXDoc0 d = let (t,m) = uncons d in case m of
  Nothing -> (t, Nothing)
  Just (p,d1) -> let (t1,m1) = deckToXDoc0 d1
                     p' = tagdocToXDoc p
                     m' = Just ([p'],cons' t1 m1)
   in case t1 of
    [] -> case m1 of Nothing -> (t, m')
                     Just (ps,d2) -> (t, Just (p' : ps, d2))
    _ -> (t, m')
tagdocToXDoc (tag,doc@(Doc d)) = case isSingletonE d of
  Nothing -> ([tag],docToXDoc doc)
  Just p -> pupd1 (tag :) $ tagdocToXDoc p

xdocToDoc (XDoc d) = Doc (xdeckToDoc d)
xdeckToDoc = feFold fe singletonF
 where
  fe d1 ps d2 = d1 `cat` foldr addP d2 ps
  addP (tags,xdoc) d = (xtagdocToDeck tags xdoc) `cat` d
  wrap tag d = singletonE (tag,Doc d)
  xtagdocToDeck ts (XDoc d) = foldr wrap (xdeckToDoc d) ts
\end{code}

\begin{code}
xdocOpt1 (XDoc d) = XDoc $ xdeckOpt1 d
xdeckOpt1 d = feFold fe singletonF d
 where
  fe d1 e d2 = ecat d1 (f e) d2
  f ps = map (pupd1 List.sort) ps

xopt1 = xdocToDoc . xdocOpt1 . docToXDoc
\end{code}

\begin{code}
consecs [] = []
consecs [l] = map (\ x -> (x,[(0,1)])) l
consecs (l:ls) = map f l ++
                 map (pupd2 (map (pupd1 succ)))
                     (filter ((`notElem` l) . fst) cs)
 where
  f y = case lookup y cs of
     Nothing -> (y,[(0,1)])
     Just ps -> let (ps0,ps') = span ((0 ==) . fst) ps
                    ps'' = map (pupd1 succ) ps'
                in case ps0 of [] -> (y, (0,1) : ps'')
                               ((_,l):_) -> (y, (0,l+1) : ps'')
  cs = consecs ls

pupd2 g (x,y) = (x, g y)
maxconsecs ls = map (pupd2 (last . List.sortBy (\ (_,x) (_,y) -> compare x y)))
                    (consecs ls)
maxconsec ls = last $ List.sortBy (\ (_,(_,x)) (_,(_,y)) -> compare x y)
                                  (maxconsecs ls)
\end{code}

\begin{code}
xdocOpt2 :: XDoc -> XDoc
xdocOpt2 (XDoc d) = XDoc $ xdeckOpt2 d

xdeckOpt2 :: XDeck -> XDeck
xdeckOpt2 d = feFold fe singletonF d
 where
  fe d1 e d2 = cat d1 $ cat (regroup (maxconsec $ map fst e) e) d2

xopt2 = xdocToDoc . xdocOpt2 . docToXDoc

xdocWrap :: [([Tag],XDoc)] -> XDeck
xdocWrap tds = singletonE $ map (pupd2 xdocOpt2) tds


regroup (tag,(offset,len)) tds = let
  (tds1,tds') = splitAt offset tds
  (tds'',tds2) = splitAt len tds'
  tds''' = map (pupd1 (filter (tag /=))) tds''
 in xdocWrap tds1 `cat`
    singletonE [([tag], XDoc $ xdocWrap tds''')] `cat`
    xdocWrap tds2
    
  
\end{code}
%}}}

\begin{code}
mirror s = case readsDoc s of
            ((d,_):_) -> show (opt1doc c_init d)
            _ -> []

opt2 = fst . opt2doc c_init

optimize = opt2 . contract1 . xopt2 . contract1 . opt2

-- optimize = contract1

smlopt2 s = case readsDoc s of
              ((d,_):_) -> show $ optimize d
              _ -> []

-- main = generateRandomDoc
-- main = interact mirror
main = interact smlopt2

d1 = "<B><I><r>bla bla </r> </I> </B> <r><I> barfoo</I></r>"
\end{code}

\section{Auxiliary Material}

\input{Parsing.lhs}

\end{document}

%{{{ EMACS lv
% Local Variables:
% folded-file: t
% fold-internal-margins: 0
% eval: (fold-set-marks "%{{{ " "%}}}")
% eval: (fold-whole-buffer)
% end:
%}}}
